@@ -199,7 +199,8 @@ double R_pow(double x, double y) /* = x ^ y */
 	return(1.);
     if(x == 0.) {
 	if(y > 0.) return(0.);
-	/* y < 0 */ return(R_PosInf);
+	else if(y < 0) return(R_PosInf);
+	else return(y); /* NA or NaN, we assert */
     }
     if (R_FINITE(x) && R_FINITE(y))
 /* work around a bug in May 2007 snapshots of gcc pre-4.3.0, also
@@ -985,8 +986,7 @@ SEXP attribute_hidden do_math1(SEXP call, SEXP op, SEXP args, SEXP env)
 
     checkArity(op, args);
 
-    if (PRIMVAL(op) != 46 &&
-	DispatchGroup("Math", call, op, args, env, &s))
+    if (DispatchGroup("Math", call, op, args, env, &s))
 	return s;
 
     if (isComplex(CAR(args)))
@@ -1026,7 +1026,7 @@ SEXP attribute_hidden do_math1(SEXP call, SEXP op, SEXP args, SEXP env)
 	   removed in 2.0.0
 	*/
 
-    case 46: return MATH1(Rf_gamma_cody);
+	/* case 46: return MATH1(Rf_gamma_cody); removed in 2.8.0 */
 
     default:
 	errorcall(call, _("unimplemented real function of 1 argument"));
@@ -1321,8 +1321,8 @@ SEXP attribute_hidden do_atan(SEXP call, SEXP op, SEXP args, SEXP env)
 /* The S4 Math2 group, round and signif */
 SEXP attribute_hidden do_Math2(SEXP call, SEXP op, SEXP args, SEXP env)
 {
-    SEXP res, ap;
-    int n, nprotect = 1;
+    SEXP res, ap, call2;
+    int n, nprotect = 2;
 
     if (length(args) >= 2 &&
 	isSymbol(CADR(args)) && R_isMissing(CADR(args), env)) {
@@ -1332,6 +1332,8 @@ SEXP attribute_hidden do_Math2(SEXP call, SEXP op, SEXP args, SEXP env)
     }
 
     PROTECT(args = evalListKeepMissing(args, env));
+    PROTECT(call2 = lang2(CAR(call), R_NilValue));
+    SETCDR(call2, args);
 
     n = length(args);
     switch (n) {
@@ -1343,7 +1345,7 @@ SEXP attribute_hidden do_Math2(SEXP call, SEXP op, SEXP args, SEXP env)
 	      n, PRIMNAME(op));
     }
 
-    if (! DispatchGroup("Math", call, op, args, env, &res)) {
+    if (! DispatchGroup("Math", call2, op, args, env, &res)) {
 	if(n == 1) {
 	    double digits = 0.0;
 	    if(PRIMVAL(op) == 10004) digits = 6.0;
@@ -1386,8 +1388,8 @@ SEXP attribute_hidden do_log1arg(SEXP call, SEXP op, SEXP args, SEXP env)
 
 SEXP attribute_hidden do_log(SEXP call, SEXP op, SEXP args, SEXP env)
 {
-    SEXP res, ap = args;
-    int n = length(args), nprotect = 1;
+    SEXP res, ap = args, call2;
+    int n = length(args), nprotect = 2;
 
     if (n >= 2 && isSymbol(CADR(args)) && R_isMissing(CADR(args), env)) {
 #ifdef M_E
@@ -1398,9 +1400,11 @@ SEXP attribute_hidden do_log(SEXP call, SEXP op, SEXP args, SEXP env)
 	PROTECT(args = list2(CAR(args), ScalarReal(e))); nprotect++;
     }
     PROTECT(args = evalListKeepMissing(args, env));
+    PROTECT(call2 = lang2(CAR(call), R_NilValue));
+    SETCDR(call2, args);
     n = length(args);
 
-    if (! DispatchGroup("Math", call, op, args, env, &res)) {
+    if (! DispatchGroup("Math", call2, op, args, env, &res)) {
 	switch (n) {
 	case 1:
 	    if (isComplex(CAR(args)))
@@ -1669,6 +1673,9 @@ SEXP attribute_hidden do_math3(SEXP call, SEXP op, SEXP args, SEXP env)
     case 43:  return Math3(args, bessel_i);
     case 44:  return Math3(args, bessel_k);
 
+    case 45:  return Math3_1(args, dnbinom_mu);
+    case 46:  return Math3_2(args, pnbinom_mu);
+    case 47:  return Math3_2(args, qnbinom_mu);
 
     default:
 	errorcall(call,
