@@ -1,7 +1,7 @@
 /*
  *  R : A Computer Language for Statistical Data Analysis
  *  Copyright (C) 1995-1998	Robert Gentleman and Ross Ihaka.
- *  Copyright (C) 2000-2005	The R Development Core Team.
+ *  Copyright (C) 2000-2007	The R Development Core Team.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -14,8 +14,8 @@
  *  GNU General Public License for more details.
  *
  *  You should have received a copy of the GNU General Public License
- *  along with this program; if not, write to the Free Software
- *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ *  along with this program; if not, a copy is available at
+ *  http://www.r-project.org/Licenses/
  *
  *
  *  print.default()  ->	 do_printdefault (with call tree below)
@@ -57,8 +57,6 @@
  *  </FIXME>
  */
 
-/* <UTF8> char here is either ASCII or handled as a whole */
-
 #ifdef HAVE_CONFIG_H
 #include <config.h>
 #endif
@@ -69,8 +67,9 @@
 #include "Rconnections.h"
 #include <S.h>
 
+
 /* Global print parameter struct: */
-R_print_par_t R_print;
+attribute_hidden R_print_par_t R_print;
 
 static void printAttributes(SEXP, SEXP, Rboolean);
 
@@ -78,6 +77,7 @@ static void printAttributes(SEXP, SEXP, Rboolean);
 static char tagbuf[TAGBUFLEN + 5];
 
 
+/* Used in X11 module for dataentry */
 void PrintDefaults(SEXP rho)
 {
     R_print.na_string = NA_STRING;
@@ -85,15 +85,18 @@ void PrintDefaults(SEXP rho)
     R_print.na_width = strlen(CHAR(R_print.na_string));
     R_print.na_width_noquote = strlen(CHAR(R_print.na_string_noquote));
     R_print.quote = 1;
-    R_print.right = 0;
+    R_print.right = Rprt_adj_left;
     R_print.digits = GetOptionDigits(rho);
     R_print.scipen = asInteger(GetOption(install("scipen"), rho));
     if (R_print.scipen == NA_INTEGER) R_print.scipen = 0;
+    R_print.max = asInteger(GetOption(install("max.print"), rho));
+    if (R_print.max == NA_INTEGER) R_print.max = 99999;
     R_print.gap = 1;
     R_print.width = GetOptionWidth(rho);
+    R_print.useSource = USESOURCE;
 }
 
-SEXP do_invisible(SEXP call, SEXP op, SEXP args, SEXP rho)
+SEXP attribute_hidden do_invisible(SEXP call, SEXP op, SEXP args, SEXP rho)
 {
     switch (length(args)) {
     case 0:
@@ -107,13 +110,14 @@ SEXP do_invisible(SEXP call, SEXP op, SEXP args, SEXP rho)
 }
 
 #if 0
-SEXP do_visibleflag(SEXP call, SEXP op, SEXP args, SEXP rho)
+SEXP attribute_hidden do_visibleflag(SEXP call, SEXP op, SEXP args, SEXP rho)
 {
     return ScalarLogical(R_Visible);
 }
 #endif
 
-SEXP do_prmatrix(SEXP call, SEXP op, SEXP args, SEXP rho)
+/* This is *only* called via outdated R_level prmatrix() : */
+SEXP attribute_hidden do_prmatrix(SEXP call, SEXP op, SEXP args, SEXP rho)
 {
     int quote;
     SEXP a, x, rowlab, collab, naprint;
@@ -127,11 +131,11 @@ SEXP do_prmatrix(SEXP call, SEXP op, SEXP args, SEXP rho)
     collab = CAR(a); a = CDR(a);
 
     quote = asInteger(CAR(a)); a = CDR(a);
-    R_print.right = asInteger(CAR(a)); a = CDR(a);
+    R_print.right = (Rprt_adj) asInteger(CAR(a)); a = CDR(a);
     naprint = CAR(a);
     if(!isNull(naprint))  {
 	if(!isString(naprint) || LENGTH(naprint) < 1)
-	    errorcall(call, _("invalid 'na.print' specification"));
+	    error(_("invalid 'na.print' specification"));
 	R_print.na_string = R_print.na_string_noquote = STRING_ELT(naprint, 0);
 	R_print.na_width = R_print.na_width_noquote =
 	    strlen(CHAR(R_print.na_string));
@@ -140,9 +144,9 @@ SEXP do_prmatrix(SEXP call, SEXP op, SEXP args, SEXP rho)
     if (length(rowlab) == 0) rowlab = R_NilValue;
     if (length(collab) == 0) collab = R_NilValue;
     if (!isNull(rowlab) && !isString(rowlab))
-	errorcall(call, _("invalid row labels"));
+	error(_("invalid row labels"));
     if (!isNull(collab) && !isString(collab))
-	errorcall(call, _("invalid column labels"));
+	error(_("invalid column labels"));
 
     printMatrix(x, 0, getAttrib(x, R_DimSymbol), quote, R_print.right,
 		rowlab, collab, rowname, colname);
@@ -150,10 +154,9 @@ SEXP do_prmatrix(SEXP call, SEXP op, SEXP args, SEXP rho)
     return x;
 }/* do_prmatrix */
 
-
 /* .Internal(print.default(x, digits, quote, na.print, print.gap,
-                           right, useS4)) */
-SEXP do_printdefault(SEXP call, SEXP op, SEXP args, SEXP rho)
+			   right, max, useS4)) */
+SEXP attribute_hidden do_printdefault(SEXP call, SEXP op, SEXP args, SEXP rho)
 {
     SEXP x, naprint;
     int tryS4;
@@ -169,19 +172,19 @@ SEXP do_printdefault(SEXP call, SEXP op, SEXP args, SEXP rho)
 	if (R_print.digits == NA_INTEGER ||
 	    R_print.digits < R_MIN_DIGITS_OPT ||
 	    R_print.digits > R_MAX_DIGITS_OPT)
-		errorcall(call, _("invalid '%s' argument"), "digits");
+		error(_("invalid '%s' argument"), "digits");
     }
     args = CDR(args);
 
     R_print.quote = asLogical(CAR(args));
     if(R_print.quote == NA_LOGICAL)
-	errorcall(call, _("invalid '%s' argument"), "quote");
+	error(_("invalid '%s' argument"), "quote");
     args = CDR(args);
 
     naprint = CAR(args);
     if(!isNull(naprint))  {
 	if(!isString(naprint) || LENGTH(naprint) < 1)
-	    errorcall(call, _("invalid 'na.print' specification"));
+	    error(_("invalid 'na.print' specification"));
 	R_print.na_string = R_print.na_string_noquote = STRING_ELT(naprint, 0);
 	R_print.na_width = R_print.na_width_noquote =
 	    strlen(CHAR(R_print.na_string));
@@ -191,40 +194,56 @@ SEXP do_printdefault(SEXP call, SEXP op, SEXP args, SEXP rho)
     if(!isNull(CAR(args))) {
 	R_print.gap = asInteger(CAR(args));
 	if (R_print.gap == NA_INTEGER || R_print.gap < 0)
-	    errorcall(call, _("'gap' must be non-negative integer"));
+	    error(_("'gap' must be non-negative integer"));
     }
     args = CDR(args);
 
-    R_print.right = asLogical(CAR(args));
+    R_print.right = (Rprt_adj) asLogical(CAR(args)); /* Should this be asInteger()? */
     if(R_print.right == NA_LOGICAL)
-	errorcall(call, _("invalid '%s' argument"), "right");
+	error(_("invalid '%s' argument"), "right");
+    args = CDR(args);
+
+    if(!isNull(CAR(args))) {
+	R_print.max = asInteger(CAR(args));
+	if(R_print.max == NA_INTEGER)
+	    error(_("invalid '%s' argument"), "max");
+    }
+    args = CDR(args);
+
+    R_print.useSource = asLogical(CAR(args));
+    if(R_print.useSource == NA_LOGICAL)
+	error(_("invalid '%s' argument"), "useSource");
+    if(R_print.useSource) R_print.useSource = USESOURCE;
     args = CDR(args);
 
     tryS4 = asLogical(CAR(args));
     if(tryS4 == NA_LOGICAL)
-	errorcall(call, _("invalid 'tryS4' internal argument"));
+	error(_("invalid 'tryS4' internal argument"));
 
-    if(tryS4 && isObject(x) && isMethodsDispatchOn()) {
-	SEXP class = getAttrib(x, R_ClassSymbol);
-	if(length(class) == 1) {
-	    /* internal version of isClass() */
-	    char str[201];
-	    snprintf(str, 200, ".__C__%s", CHAR(STRING_ELT(class, 0)));
-	    if(findVar(install(str), rho) != R_UnboundValue)
-		callShow = TRUE;
-	}
-    }
+    if(tryS4 && IS_S4_OBJECT(x) && isMethodsDispatchOn())
+	callShow = TRUE;
 
     if(callShow) {
-	SEXP call;
-	PROTECT(call = lang2(install("show"), x));
+	/* we need to get show from the methods namespace if it is
+	   not visible on the search path. */
+	SEXP call, showS;
+	showS = findVar(install("show"), rho);
+	if(showS == R_UnboundValue) {
+	    SEXP methodsNS = R_FindNamespace(mkString("methods"));
+	    if(methodsNS == R_UnboundValue)
+		error("missing methods namespace: this should not happen");
+	    showS = findVarInFrame3(methodsNS, install("show"), TRUE);
+	    if(showS == R_UnboundValue)
+		error("missing show() in methods namespace: this should not happen");
+	}
+	PROTECT(call = lang2(showS, x));
 	eval(call, rho);
 	UNPROTECT(1);
     } else {
 	CustomPrintValue(x, rho);
     }
 
-    PrintDefaults(rho); /* reset, as na.print.etc may have been set */
+    PrintDefaults(rho); /* reset, as na.print etc may have been set */
     return x;
 }/* do_printdefault */
 
@@ -237,12 +256,13 @@ static void PrintGenericVector(SEXP s, SEXP env)
 {
     int i, taglen, ns, w, d, e, wr, dr, er, wi, di, ei;
     SEXP dims, t, names, newcall, tmp;
-    char pbuf[115], *ptag, *rn, *cn, save[TAGBUFLEN + 5];
+    char pbuf[115], *ptag, save[TAGBUFLEN + 5];
 
     ns = length(s);
     if((dims = getAttrib(s, R_DimSymbol)) != R_NilValue && length(dims) > 1) {
 	PROTECT(dims);
 	PROTECT(t = allocArray(STRSXP, dims));
+	/* FIXME: check (ns <= R_print.max +1) ? ns : R_print.max; */
 	for (i = 0 ; i < ns ; i++) {
 	    switch(TYPEOF(PROTECT(tmp = VECTOR_ELT(s, i)))) {
 	    case NILSXP:
@@ -251,7 +271,7 @@ static void PrintGenericVector(SEXP s, SEXP env)
 	    case LGLSXP:
 		if (LENGTH(tmp) == 1) {
 		    formatLogical(LOGICAL(tmp), 1, &w);
-		    snprintf(pbuf, 115, "%s", 
+		    snprintf(pbuf, 115, "%s",
 			     EncodeLogical(LOGICAL(tmp)[0], w));
 		} else
 		    snprintf(pbuf, 115, "Logical,%d", LENGTH(tmp));
@@ -272,7 +292,7 @@ static void PrintGenericVector(SEXP s, SEXP env)
 	    case REALSXP:
 		if (LENGTH(tmp) == 1) {
 		    formatReal(REAL(tmp), 1, &w, &d, &e, 0);
-		    snprintf(pbuf, 115, "%s", 
+		    snprintf(pbuf, 115, "%s",
 			     EncodeReal(REAL(tmp)[0], w, d, e, OutDec));
 		} else
 		    snprintf(pbuf, 115, "Numeric,%d", LENGTH(tmp));
@@ -282,7 +302,7 @@ static void PrintGenericVector(SEXP s, SEXP env)
 		    Rcomplex *x = COMPLEX(tmp);
 		    formatComplex(x, 1, &wr, &dr, &er, &wi, &di, &ei, 0);
 		    if (ISNA(x[0].r) || ISNA(x[0].i))
-			snprintf(pbuf, 115, "%s", 
+			snprintf(pbuf, 115, "%s",
 				 EncodeReal(NA_REAL, w, 0, 0, OutDec));
 		    else
 			snprintf(pbuf, 115, "%s", EncodeComplex(x[0],
@@ -293,7 +313,7 @@ static void PrintGenericVector(SEXP s, SEXP env)
 	    case STRSXP:
 		if (LENGTH(tmp) == 1) {
 		    /* This can potentially overflow */
-		    char *ctmp = CHAR(STRING_ELT(tmp, 0));
+		    const char *ctmp = translateChar(STRING_ELT(tmp, 0));
 		    int len = strlen(ctmp);
 		    if(len < 100)
 			snprintf(pbuf, 115, "\"%s\"", ctmp);
@@ -325,6 +345,7 @@ static void PrintGenericVector(SEXP s, SEXP env)
 	}
 	if (LENGTH(dims) == 2) {
 	    SEXP rl, cl;
+	    const char *rn, *cn;
 	    GetMatrixDimnames(s, &rl, &cl, &rn, &cn);
 	    /* as from 1.5.0: don't quote here as didn't in array case */
 	    printMatrix(t, 0, dims, 0, R_print.right, rl, cl,
@@ -332,11 +353,11 @@ static void PrintGenericVector(SEXP s, SEXP env)
 	}
 	else {
 	    names = GetArrayDimnames(s);
-	    printArray(t, dims, 0, 0, names);
+	    printArray(t, dims, 0, Rprt_adj_left, names);
 	}
 	UNPROTECT(2);
     }
-    else {
+    else { /* .. no dim() .. */
 	names = getAttrib(s, R_NamesSymbol);
 	taglen = strlen(tagbuf);
 	ptag = tagbuf + taglen;
@@ -345,55 +366,62 @@ static void PrintGenericVector(SEXP s, SEXP env)
 	SET_TYPEOF(newcall, LANGSXP);
 
 	if(ns > 0) {
-	for (i = 0 ; i < ns ; i++) {
-	    if (i > 0) Rprintf("\n");
-	    if (names != R_NilValue &&
-		STRING_ELT(names, i) != R_NilValue &&
-		*CHAR(STRING_ELT(names, i)) != '\0') {
-		if (taglen + strlen(CHAR(STRING_ELT(names, i))) > TAGBUFLEN)
-		    sprintf(ptag, "$...");
-		else {
-		    /* we need to distinguish character NA from "NA", which
-		       is a valid (if non-syntactic) name */
-		    if (STRING_ELT(names, i) == NA_STRING)
-			sprintf(ptag, "$<NA>");
-		    else if( isValidName(CHAR(STRING_ELT(names, i))) )
-			sprintf(ptag, "$%s", CHAR(STRING_ELT(names, i)));
-		    else
-			sprintf(ptag, "$\"%s\"", CHAR(STRING_ELT(names, i)));
+	    int n_pr = (ns <= R_print.max +1) ? ns : R_print.max;
+	    /* '...max +1'  ==> will omit at least 2 ==> plural in msg below */
+	    for (i = 0 ; i < n_pr ; i++) {
+		if (i > 0) Rprintf("\n");
+		if (names != R_NilValue &&
+		    STRING_ELT(names, i) != R_NilValue &&
+		    *CHAR(STRING_ELT(names, i)) != '\0') {
+		    const char *ss = translateChar(STRING_ELT(names, i));
+		    if (taglen + strlen(ss) > TAGBUFLEN)
+			sprintf(ptag, "$...");
+		    else {
+			/* we need to distinguish character NA from "NA", which
+			   is a valid (if non-syntactic) name */
+			if (STRING_ELT(names, i) == NA_STRING)
+			    sprintf(ptag, "$<NA>");
+			else if( isValidName(ss) )
+			    sprintf(ptag, "$%s", ss);
+			else
+			    sprintf(ptag, "$`%s`", ss);
+		    }
 		}
+		else {
+		    if (taglen + IndexWidth(i) > TAGBUFLEN)
+			sprintf(ptag, "$...");
+		    else
+			sprintf(ptag, "[[%d]]", i+1);
+		}
+		Rprintf("%s\n", tagbuf);
+		if(isObject(VECTOR_ELT(s, i))) {
+		    /* need to preserve tagbuf */
+		    strcpy(save, tagbuf);
+		    SETCADR(newcall, VECTOR_ELT(s, i));
+		    eval(newcall, env);
+		    strcpy(tagbuf, save);
+		}
+		else PrintValueRec(VECTOR_ELT(s, i), env);
+		*ptag = '\0';
 	    }
-	    else {
-		if (taglen + IndexWidth(i) > TAGBUFLEN)
-		    sprintf(ptag, "$...");
-		else
-		    sprintf(ptag, "[[%d]]", i+1);
-	    }
-	    Rprintf("%s\n", tagbuf);
-	    if(isObject(VECTOR_ELT(s, i))) {
-		/* need to preserve tagbuf */
-		strcpy(save, tagbuf);
-		SETCADR(newcall, VECTOR_ELT(s, i));
-		eval(newcall, env);
-		strcpy(tagbuf, save);
-	    }
-	    else PrintValueRec(VECTOR_ELT(s, i), env);
-	    *ptag = '\0';
+	    Rprintf("\n");
+	    if(n_pr < ns)
+		Rprintf(" [ reached getOption(\"max.print\") -- omitted %d entries ]]\n",
+			ns - n_pr);
 	}
-	Rprintf("\n");
-	}
-	else {
+	else { /* ns = length(s) == 0 */
 	    /* Formal classes are represented as empty lists */
-	    char *className = NULL;
-	    SEXP class;
+	    const char *className = NULL;
+	    SEXP klass;
 	    if(isObject(s) && isMethodsDispatchOn()) {
-		class = getAttrib(s, R_ClassSymbol);
-		if(length(class) == 1) {
+		klass = getAttrib(s, R_ClassSymbol);
+		if(length(klass) == 1) {
 		    /* internal version of isClass() */
 		    char str[201];
-		    snprintf(str, 200, ".__C__%s", CHAR(STRING_ELT(class, 0)));
+		    const char *ss = translateChar(STRING_ELT(klass, 0));
+		    snprintf(str, 200, ".__C__%s", ss);
 		    if(findVar(install(str), env) != R_UnboundValue)
-			className = CHAR(STRING_ELT(class, 0));
+			className = ss;
 		}
 	    }
 	    if(className) {
@@ -414,7 +442,8 @@ static void printList(SEXP s, SEXP env)
 {
     int i, taglen;
     SEXP dims, dimnames, t, newcall;
-    char pbuf[101], *ptag, *rn, *cn;
+    char pbuf[101], *ptag;
+    const char *rn, *cn;
 
     if ((dims = getAttrib(s, R_DimSymbol)) != R_NilValue && length(dims) > 1) {
 	PROTECT(dims);
@@ -472,7 +501,7 @@ static void printList(SEXP s, SEXP env)
 	}
 	else {
 	    dimnames = getAttrib(s, R_DimNamesSymbol);
-	    printArray(t, dims, 0, 0, dimnames);
+	    printArray(t, dims, 0, Rprt_adj_left, dimnames);
 	}
 	UNPROTECT(2);
     }
@@ -496,7 +525,7 @@ static void printList(SEXP s, SEXP env)
 		    else if( isValidName(CHAR(PRINTNAME(TAG(s)))) )
 			sprintf(ptag, "$%s", CHAR(PRINTNAME(TAG(s))));
 		    else
-			sprintf(ptag, "$\"%s\"", CHAR(PRINTNAME(TAG(s))));
+			sprintf(ptag, "$`%s`", CHAR(PRINTNAME(TAG(s))));
 		}
 	    }
 	    else {
@@ -530,38 +559,45 @@ static void PrintExpression(SEXP s)
     SEXP u;
     int i, n;
 
-    u = deparse1(s, 0, SIMPLEDEPARSE);
+    u = deparse1(s, 0, R_print.useSource | DEFAULTDEPARSE);
     n = LENGTH(u);
     for (i = 0; i < n ; i++)
-	Rprintf("%s\n", CHAR(STRING_ELT(u, i)));
+	Rprintf("%s\n", CHAR(STRING_ELT(u, i))); /*translated */
 }
 
-
 /* PrintValueRec -- recursively print an SEXP
 
  * This is the "dispatching" function for  print.default()
  */
-
-static void PrintEnvir(SEXP rho)
-{
-    if (rho == R_GlobalEnv)
-	Rprintf("<environment: R_GlobalEnv>\n");
-    else if (rho == R_BaseEnv)
-    	Rprintf("<environment: base>\n");	
-    else if (R_IsPackageEnv(rho))
-	Rprintf("<environment: %s>\n",
-		CHAR(STRING_ELT(R_PackageEnvName(rho), 0)));
-    else if (R_IsNamespaceEnv(rho))
-	Rprintf("<environment: namespace:%s>\n",
-		CHAR(STRING_ELT(R_NamespaceEnvSpec(rho), 0)));
-    else Rprintf("<environment: %p>\n", rho);
-}
-
-void PrintValueRec(SEXP s,SEXP env)
+void attribute_hidden PrintValueRec(SEXP s, SEXP env)
 {
     int i;
     SEXP t;
 
+#ifdef Win32
+    WinCheckUTF8();
+#endif
+    if(!isMethodsDispatchOn() && (IS_S4_OBJECT(s) || TYPEOF(s) == S4SXP) ) {
+	SEXP cl = getAttrib(s, install("class"));
+	if(isNull(cl)) {
+	    /* This might be a mistaken S4 bit set */
+	    if(TYPEOF(s) == S4SXP)
+		Rprintf("<S4 object without a class>\n");
+	    else
+		Rprintf("<Object of type '%s' with S4 bit but without a class>\n",
+			type2char(TYPEOF(s)));
+	} else {
+	    SEXP pkg = getAttrib(s, install("package"));
+	    if(isNull(pkg)) {
+		Rprintf("<S4 object of class \"%s\">\n",
+			CHAR(STRING_ELT(cl, 0)));
+	    } else {
+		Rprintf("<S4 object of class \"%s\" from package '%s'>\n",
+			CHAR(STRING_ELT(cl, 0)), CHAR(STRING_ELT(pkg, 0)));
+	    }
+	}
+	return;
+    }
     switch (TYPEOF(s)) {
     case NILSXP:
 	Rprintf("NULL\n");
@@ -569,15 +605,42 @@ void PrintValueRec(SEXP s,SEXP env)
     case SYMSXP: /* Use deparse here to handle backtick quotification
 		  * of "weird names" */
 	t = deparse1(s, 0, SIMPLEDEPARSE);
-	Rprintf("%s\n", CHAR(STRING_ELT(t, 0)));
+	Rprintf("%s\n", CHAR(STRING_ELT(t, 0))); /* translated */
 	break;
     case SPECIALSXP:
     case BUILTINSXP:
-	Rprintf(".Primitive(\"%s\")\n", PRIMNAME(s));
+	/* This is OK as .Internals are not visible to be printed */
+    {
+	char *nm = PRIMNAME(s);
+	SEXP env, s2;
+	PROTECT_INDEX xp;
+	PROTECT_WITH_INDEX(env = findVarInFrame3(R_BaseEnv,
+						 install(".ArgsEnv"), TRUE),
+			   &xp);
+	if (TYPEOF(env) == PROMSXP) REPROTECT(env = eval(env, R_BaseEnv), xp);
+	s2 = findVarInFrame3(env, install(nm), TRUE);
+	if(s2 == R_UnboundValue) {
+	    REPROTECT(env = findVarInFrame3(R_BaseEnv,
+					    install(".GenericArgsEnv"), TRUE),
+		      xp);
+	    if (TYPEOF(env) == PROMSXP)
+		REPROTECT(env = eval(env, R_BaseEnv), xp);
+	    s2 = findVarInFrame3(env, install(nm), TRUE);
+	}
+	if(s2 != R_UnboundValue) {
+	    PROTECT(s2);
+	    t = deparse1(s2, 0, DEFAULTDEPARSE);
+	    Rprintf("%s ", CHAR(STRING_ELT(t, 0))); /* translated */
+	    Rprintf(".Primitive(\"%s\")\n", PRIMNAME(s));
+	    UNPROTECT(1);
+	} else /* missing definition, e.g. 'if' */
+	    Rprintf(".Primitive(\"%s\")\n", PRIMNAME(s));
+	UNPROTECT(1);
+    }
 	break;
     case CHARSXP:
 	Rprintf("<CHARSXP: ");
-	Rprintf(EncodeString(s, 0, '"', Rprt_adj_left));
+	Rprintf("%s", EncodeString(s, 0, '"', Rprt_adj_left));
 	Rprintf(">\n");
 	break;
     case EXPRSXP:
@@ -586,25 +649,26 @@ void PrintValueRec(SEXP s,SEXP env)
     case CLOSXP:
     case LANGSXP:
 	t = getAttrib(s, R_SourceSymbol);
-	if (isNull(t))
-	    t = deparse1(s, 0, SIMPLEDEPARSE);
+	if (!isString(t) || !R_print.useSource)
+	    t = deparse1(s, 0, R_print.useSource | DEFAULTDEPARSE);
 	for (i = 0; i < LENGTH(t); i++)
-	    Rprintf("%s\n", CHAR(STRING_ELT(t, i)));
+	    Rprintf("%s\n", CHAR(STRING_ELT(t, i))); /* translated */
 #ifdef BYTECODE
 	if (TYPEOF(s) == CLOSXP && isByteCode(BODY(s)))
 	    Rprintf("<bytecode: %p>\n", BODY(s));
 #endif
-	if (TYPEOF(s) == CLOSXP) t = CLOENV(s);
-	else t = R_GlobalEnv;
-	if (t != R_GlobalEnv)
-	    PrintEnvir(t);
+	if (TYPEOF(s) == CLOSXP) {
+	    t = CLOENV(s);
+	    if (t != R_GlobalEnv)
+		Rprintf("%s\n", EncodeEnvironment(t));
+	}
 	break;
     case ENVSXP:
-	PrintEnvir(s);
+	Rprintf("%s\n", EncodeEnvironment(s));
 	break;
     case PROMSXP:
-        Rprintf("<promise: %p>\n", s);
-        break;
+	Rprintf("<promise: %p>\n", s);
+	break;
     case DOTSXP:
 	Rprintf("<...>\n");
 	break;
@@ -626,10 +690,10 @@ void PrintValueRec(SEXP s,SEXP env)
 		PROTECT(t = getAttrib(s, R_DimNamesSymbol));
 		if (t != R_NilValue && VECTOR_ELT(t, 0) != R_NilValue) {
 		    SEXP nn = getAttrib(t, R_NamesSymbol);
-		    char *title = NULL;
+		    const char *title = NULL;
 
 		    if (!isNull(nn))
-		        title = CHAR(STRING_ELT(nn, 0));
+			title = translateChar(STRING_ELT(nn, 0));
 
 		    printNamedVector(s, VECTOR_ELT(t, 0), R_print.quote, title);
 		}
@@ -639,7 +703,7 @@ void PrintValueRec(SEXP s,SEXP env)
 	    }
 	    else if (LENGTH(t) == 2) {
 		SEXP rl, cl;
-		char *rn, *cn;
+		const char *rn, *cn;
 		GetMatrixDimnames(s, &rl, &cl, &rn, &cn);
 		printMatrix(s, 0, t, R_print.quote, R_print.right, rl, cl,
 			    rn, cn);
@@ -671,10 +735,19 @@ void PrintValueRec(SEXP s,SEXP env)
     case WEAKREFSXP:
 	Rprintf("<weak reference>\n");
 	break;
+    case S4SXP:
+	/*  we got here because no show method, usually no class.
+	    Print the "slots" as attributes, since we don't know the class.
+	*/
+	Rprintf("<S4 Type Object>\n");
+	break;
     default:
 	UNIMPLEMENTED_TYPE("PrintValueRec", s);
     }
     printAttributes(s, env, FALSE);
+#ifdef Win32
+    WinUTF8out = FALSE;
+#endif
 }
 
 /* 2000-12-30 PR#715: remove list tags from tagbuf here
@@ -703,7 +776,7 @@ static void printAttributes(SEXP s, SEXP env, Rboolean useSlots)
 		   TAG(a) == R_DimNamesSymbol)
 		    goto nextattr;
 	    }
-	    if(isFactor(s)) {
+	    if(inherits(s, "factor")) {
 		if(TAG(a) == R_LevelsSymbol)
 		    goto nextattr;
 		if(TAG(a) == R_ClassSymbol)
@@ -717,7 +790,7 @@ static void printAttributes(SEXP s, SEXP env, Rboolean useSlots)
 		if (TAG(a) == R_NamesSymbol)
 		    goto nextattr;
 	    }
-	    if(TAG(a) == R_CommentSymbol || TAG(a) == R_SourceSymbol)
+	    if(TAG(a) == R_CommentSymbol || TAG(a) == R_SourceSymbol || TAG(a) == R_SrcrefSymbol)
 		goto nextattr;
 	    if(useSlots)
 		sprintf(ptag, "Slot \"%s\":",
@@ -726,7 +799,30 @@ static void printAttributes(SEXP s, SEXP env, Rboolean useSlots)
 		sprintf(ptag, "attr(,\"%s\")",
 			EncodeString(PRINTNAME(TAG(a)), 0, 0, Rprt_adj_left));
 	    Rprintf("%s", tagbuf); Rprintf("\n");
-	    if (isObject(CAR(a))) {
+	    if (TAG(a) == R_RowNamesSymbol) {
+		/* need special handling AND protection */
+		SEXP val;
+		PROTECT(val = getAttrib(s, R_RowNamesSymbol));
+		PrintValueRec(val, env);
+		UNPROTECT(1);
+		goto nextattr;
+	    }
+	    if (isMethodsDispatchOn() && IS_S4_OBJECT(CAR(a))) {
+		SEXP s, showS;
+
+		showS = findVar(install("show"), env);
+		if(showS == R_UnboundValue) {
+		    SEXP methodsNS = R_FindNamespace(mkString("methods"));
+		    if(methodsNS == R_UnboundValue)
+			error("missing methods namespace: this should not happen");
+		    showS = findVarInFrame3(methodsNS, install("show"), TRUE);
+		    if(showS == R_UnboundValue)
+			error("missing show() in methods namespace: this should not happen");
+		}
+		PROTECT(s = lang2(showS, CAR(a)));
+		eval(s, env);
+		UNPROTECT(1);
+	    } else if (isObject(CAR(a))) {
 		/* Need to construct a call to
 		   print(CAR(a), digits)
 		   based on the R_print structure, then eval(call, env).
@@ -734,29 +830,23 @@ static void printAttributes(SEXP s, SEXP env, Rboolean useSlots)
 
 		   quote, right, gap should probably be included if
 		   they have non-missing values.
+
+		   This will not dispatch to show() as 'digits' is supplied.
 		*/
 		SEXP s, t, na_string = R_print.na_string,
 		    na_string_noquote = R_print.na_string_noquote;
-		int quote = R_print.quote, right = R_print.right,
+		int quote = R_print.quote,
 		    digits = R_print.digits, gap = R_print.gap,
 		    na_width = R_print.na_width,
 		    na_width_noquote = R_print.na_width_noquote;
+		Rprt_adj right = R_print.right;
+
 		PROTECT(t = s = allocList(3));
 		SET_TYPEOF(s, LANGSXP);
 		SETCAR(t, install("print")); t = CDR(t);
 		SETCAR(t,  CAR(a)); t = CDR(t);
-		SETCAR(t, allocVector(INTSXP, 1));
-		INTEGER(CAR(t))[0] = digits;
-		SET_TAG(t, install("digits")); /* t = CDR(t);
-		SETCAR(t, allocVector(LGLSXP, 1));
-		LOGICAL(CAR(t))[0] = quote;
-		SET_TAG(t, install("quote")); t = CDR(t);
-		SETCAR(t, allocVector(LGLSXP, 1));
-		LOGICAL(CAR(t))[0] = right;
-		SET_TAG(t, install("right")); t = CDR(t);
-		SETCAR(t, allocVector(INTSXP, 1));
-		INTEGER(CAR(t))[0] = gap;
-		SET_TAG(t, install("gap")); */
+		SETCAR(t, ScalarInteger(digits));
+		SET_TAG(t, install("digits"));
 		eval(s, env);
 		UNPROTECT(1);
 		R_print.quote = quote;
@@ -779,37 +869,43 @@ static void printAttributes(SEXP s, SEXP env, Rboolean useSlots)
 
 
 /* Print an S-expression using (possibly) local options.
-   This is used for auto-printing */
+   This is used for auto-printing from main.c */
 
-void PrintValueEnv(SEXP s, SEXP env)
+void attribute_hidden PrintValueEnv(SEXP s, SEXP env)
 {
-    SEXP call;
-    char *autoprint = "print";
-
     PrintDefaults(env);
     tagbuf[0] = '\0';
     PROTECT(s);
     if(isObject(s)) {
-	/* The intention here is to call show() on S4 objects, otherwise
+	/*
+	   The intention here is to call show() on S4 objects, otherwise
 	   print(), so S4 methods for show() have precedence over those for
-	   print(). We decided not to do this, at least for now.
+	   print() to conform with the "green book", p. 332
 	*/
-        /*if(isMethodsDispatchOn()) {
-	    SEXP class = getAttrib(s, R_ClassSymbol);
-	    if(length(class) == 1) {
-		char str[201];
-		snprintf(str, 200, ".__C__%s", CHAR(STRING_ELT(class, 0)));
-		if(findVar(install(str), env) != R_UnboundValue)
-		    autoprint = "show";
+	SEXP call, showS;
+	if(isMethodsDispatchOn() && IS_S4_OBJECT(s)) {
+	    /*
+	      Note that we cannot assume that show() is visible from
+	      'env', but we can assume there is a loaded "methods"
+	      namespace.  It is tempting to cache the value of show in
+	      the namespace, but the latter could be unloaded and
+	      reloaded in a session.
+	    */
+	    showS = findVar(install("show"), env);
+	    if(showS == R_UnboundValue) {
+		SEXP methodsNS = R_FindNamespace(mkString("methods"));
+		if(methodsNS == R_UnboundValue)
+		    error("missing methods namespace: this should not happen");
+		showS = findVarInFrame3(methodsNS, install("show"), TRUE);
+		if(showS == R_UnboundValue)
+		    error("missing show() in methods namespace: this should not happen");
 	    }
-	}*/
-	PROTECT(call = lang2(install(autoprint), s));
+	    PROTECT(call = lang2(showS, s));
+	} else
+	    PROTECT(call = lang2(install("print"), s));
 	eval(call, env);
 	UNPROTECT(1);
-    }
-    else {
-	PrintValueRec(s, env);
-    }
+    } else PrintValueRec(s, env);
     UNPROTECT(1);
 }
 
@@ -818,7 +914,7 @@ void PrintValueEnv(SEXP s, SEXP env)
 
 void PrintValue(SEXP s)
 {
-    PrintValueEnv(s, R_BaseEnv);
+    PrintValueEnv(s, R_GlobalEnv);
 }
 
 
@@ -826,20 +922,23 @@ void PrintValue(SEXP s)
 
 void R_PV(SEXP s)
 {
-    if(isObject(s)) PrintValueEnv(s, R_BaseEnv);
+    if(isObject(s)) PrintValueEnv(s, R_GlobalEnv);
 }
 
 
-void CustomPrintValue(SEXP s, SEXP env)
+void attribute_hidden CustomPrintValue(SEXP s, SEXP env)
 {
     tagbuf[0] = '\0';
     PrintValueRec(s, env);
 }
 
 
-/* xxxpr are mostly for S compatibility (as mentioned in V&R) */
+/* xxxpr are mostly for S compatibility (as mentioned in V&R).
+   The actual interfaces are now in xxxpr.f
+ */
 
-int F77_NAME(dblep0) (char *label, int *nchar, double *data, int *ndata)
+attribute_hidden
+int F77_NAME(dblep0) (const char *label, int *nchar, double *data, int *ndata)
 {
     int k, nc = *nchar;
 
@@ -856,7 +955,8 @@ int F77_NAME(dblep0) (char *label, int *nchar, double *data, int *ndata)
     return(0);
 }
 
-int F77_NAME(intpr0) (char *label, int *nchar, int *data, int *ndata)
+attribute_hidden
+int F77_NAME(intpr0) (const char *label, int *nchar, int *data, int *ndata)
 {
     int k, nc = *nchar;
 
@@ -873,9 +973,10 @@ int F77_NAME(intpr0) (char *label, int *nchar, int *data, int *ndata)
     return(0);
 }
 
-int F77_NAME(realp0) (char *label, int *nchar, float *data, int *ndata)
+attribute_hidden
+int F77_NAME(realp0) (const char *label, int *nchar, float *data, int *ndata)
 {
-    int k, nc = *nchar, nd=*ndata;
+    int k, nc = *nchar, nd = *ndata;
     double *ddata;
 
     if(nc < 0) nc = strlen(label);
@@ -889,7 +990,7 @@ int F77_NAME(realp0) (char *label, int *nchar, float *data, int *ndata)
 	Rprintf("\n");
     }
     if(nd > 0) {
-	ddata = malloc(nd*sizeof(double));
+	ddata = (double *) malloc(nd*sizeof(double));
 	if(!ddata) error(_("memory allocation error in realpr"));
 	for (k = 0; k < nd; k++) ddata[k] = (double) data[k];
 	printRealVector(ddata, nd, 1);
@@ -900,7 +1001,7 @@ int F77_NAME(realp0) (char *label, int *nchar, float *data, int *ndata)
 
 /* Fortran-callable error routine for lapack */
 
-void F77_NAME(xerbla)(char *srname, int *info)
+void F77_NAME(xerbla)(const char *srname, int *info)
 {
    /* srname is not null-terminated.  It should be 6 characters. */
     char buf[7];
